# This function computes the conditional estimates
# using bootstrap for an object of gmnl for latent class models
#
# Return a list with two elements: 1) a matrix n * k with the mean
# 2) a matrix n*k with the se
#

boot.lc <- function(object, panel = FALSE, S = 2, Q = 2, reflevel = NULL){
  if (!inherits(object, "gmnl")) stop("not a \"gmnl\" object")
  model <- object$model
  if (model != "lc") stop("Just lc model allowed for now")
  
  # Extract terms
  formula <- object$formula
  mf <- object$mf
  
  
  index <- attr(mf, "index")
  alt   <- index[["alt"]]
  chid  <- index[["chid"]]
  
  alt.lev <- levels(alt)
  J <- length(alt.lev)
  
  ## Check Panel
  if (panel) {
    #if (model == "mnl") stop("Panel is not relevant for mnl model")
    id <- index[["id"]]
    if (is.null(id)) stop("No individual index")
    id <- split(index[["id"]], alt)[[1]]
  } else{
    id <- NULL
  }
  
  # Weights
  if (any(names(mf) == "(weights)")) {
    weights <- mf[["(weights)"]] <- mf[["(weights)"]]/mean(mf[["(weights)"]])
    weights <- split(weights, alt)[[1]]
  } else weights <- 1
  
  ## Data
  y.name <- paste(deparse(object$formula[[2]]))
  y <- object$mf[[y.name]]
  X <- model.matrix(object$formula, object$mf)
  
  # List of Variables
  Xl <- vector(length = J, mode = "list")
  names(Xl) <- levels(alt)
  for (i in levels(alt)) {
    Xl[[i]] <- X[alt == i, , drop = FALSE]
  }
  yl <- split(y, alt)
  yl <- lapply(yl, function(x){x[is.na(x)] <- FALSE ; x})
  H <- gmnl:::model.matrix.gFormula(formula, mf, rhs = 5, Q = Q)
  classes <- attr(H, "alt")
  Hl <- vector(length = Q, mode = "list")
  names(Hl) <- levels(classes)
  for (i in levels(classes)) {
    Hl[[i]] <- H[classes == i, , drop = FALSE]
  }
  ## Generate theta.hat
  beta.hat <- coef(object)
  Sigma.hat <- vcov(object)
  L <- t(chol(Sigma.hat))
  # FIXME: probably we should fixed the seed for eta
  eta <- matrix(rnorm(length(beta.hat) * S), nrow = length(beta.hat), ncol = S)
  theta.hat <- beta.hat  + L %*% eta
  
  # array to save restuls from boot
  mean.boot <- array(NA, dim = c(ifelse(panel, length(unique(id)), nrow(Xl[[1]])), ncol(Xl[[1]]), S))
  
  # Loop for bootstraping
  for (s in 1:S) {
    out <- gmnl:::ll.mlogitlc(theta =  theta.hat[, s], 
                              y = yl, 
                              X = Xl, 
                              H = Hl, 
                              Q = Q, 
                              id = id,
                              weights = weights,
                              gradient =  FALSE,
                              get.bi =  TRUE)
    
    # Compute conditional estimates
    bi <- attr(out, 'bi')
    Qir <- attr(out, 'Qir')
    N  <- nrow(Qir)
    K <- ncol(bi)
    
    mean <- array(NA, dim = c(N, Q, K))
    for (j in 1:K) mean[, , j] <- gmnl:::repRows(bi[, j], N)   * Qir
    mean.boot[,, s] <- apply(mean,  c(1, 3), sum)
  }
  
  names <- dimnames(bi)[[2]]
  dimnames(mean.boot) <- list(NULL, names, NULL)
  meanCb <- apply(mean.boot, c(1, 2), mean)
  sdCb   <- apply(mean.boot, c(1, 2), sd)
  quant  <- apply(mean.boot, c(1, 2), quantile, probs = c(0.025, 0.975))
  boot.r <- list(
              mean = meanCb,
              se = sdCb,
              quant = quant)
  return(boot.r)
}